// -*- C++ -*-
//
//  This file is part of Siena, a wide-area event notification system.
//  See http://www.cs.colorado.edu/serl/siena/
//
//  Authors: See the file AUTHORS for full details.
//
//  Copyright (C) 2002-2004 University of Colorado
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307,
//  USA, or send email to serl@cs.colorado.edu.
//
// $Id: simplify.h,v 1.2 2008-08-21 15:34:10 carzanig Exp $
//
#ifndef SIENA_SIMPLIFY_H
#define SIENA_SIMPLIFY_H

#include "siena/types.h"

/** \file simplify.h 
 *
 *  This header file defines the functions that implement the Siena
 *  predicate simplification logic.
 **/

/** \namespace siena
 *
 *  @brief name space for Siena and related types and modules.
 **/
namespace siena
{
  /** @brief Procedure to determine if two constraints conflict with each other.
   *
   * Constraints conflict if they can never be true for the same
   * message.  Conflicting constraints in the same filter will render
   * the filter unsatisfiable, so the filter can effectively be
   * eliminated from consideration.
   * <p>
   * Constraints <b>never</b> conflict if:
   * <ul>
   * <li> They are they same object (addresses are equal).
   * <li> Their names are not equal.
   * </ul>
   * Otherwise, constraints conflict if:
   * <ul>
   * <li> They have different types (since a message attribute only has a single type).
   * <li> According to this table:
   *  <table>
   *   <tr>
   *    <td>&nbsp;</td>
   *    <td><code><b>x = b</b></code></td>
   *    <td><code><b>x < b</b></code></td>
   *    <td><code><b>x > b</b></code></td>
   *    <td><code><b>x *= b</b></code></td>
   *    <td><code><b>x =* b</b></code></td>
   *    <td><code><b>x ** b</b></code></td>
   *    <td><code><b>x any any</b></code></td>
   *    <td><code><b>x != b</b></code></td>
   *   </tr>
   *   <tr>
   *    <td><code><b>x = a</b></code></td>
   *    <td><code>if b != a</code></td>
   *    <td><code>if b <= a</code></td>
   *    <td><code>if b >= a</code></td>
   *    <td><code>if !(b *= a)</code></td>
   *    <td><code>if !(b =* a)</code></td>
   *    <td><code>if !(b ** a)</code></td>
   *    <td><code>false</code></td>
   *    <td><code>if b = a</code></td>
   *   </tr>
   *   <tr>
   *    <td><code><b>x < a</b></code></td>
   *    <td><code>if b >= a</code></td>
   *    <td><code>false</code></td>
   *    <td><code>if b >= a</code></td>
   *    <td><code>false</code></td>
   *    <td><code>false</code></td>
   *    <td><code>false</code></td>
   *    <td><code>false</code></td>
   *    <td><code>false</code></td>
   *   </tr>
   *   <tr>
   *    <td><code><b>x > a</b></code></td>
   *    <td><code>if b <= a</code></td>
   *    <td><code>if b <= a</code></td>
   *    <td><code>false</code></td>
   *    <td><code>false</code></td>
   *    <td><code>false</code></td>
   *    <td><code>false</code></td>
   *    <td><code>false</code></td>
   *    <td><code>false</code></td>
   *   </tr>
   *   <tr>
   *    <td><code><b>x *= a</b></code></td>
   *    <td><code>if !(a *= b)</code></td>
   *    <td><code>false</code></td>
   *    <td><code>false</code></td>
   *    <td><code>if !(a *= b)<br>&&<br>!(b *= a)</code></td>
   *    <td><code>false</code></td>
   *    <td><code>false</code></td>
   *    <td><code>false</code></td>
   *    <td><code>false</code></td>
   *   </tr>
   *   <tr>
   *    <td><code><b>x =* a</b></code></td>
   *    <td><code>if !(a =* b)</code></td>
   *    <td><code>false</code></td>
   *    <td><code>false</code></td>
   *    <td><code>false</code></td>
   *    <td><code>if !(a =* b)<br>&&<br>!(b =* a)</code></td>
   *    <td><code>false</code></td>
   *    <td><code>false</code></td>
   *    <td><code>false</code></td>
   *   </tr>
   *   <tr>
   *    <td><code><b>x ** a</b></code></td>
   *    <td><code>if !(a ** b)</code></td>
   *    <td><code>false</code></td>
   *    <td><code>false</code></td>
   *    <td><code>false</code></td>
   *    <td><code>false</code></td>
   *    <td><code>false</code></td>
   *    <td><code>false</code></td>
   *    <td><code>false</code></td>
   *   </tr>
   *   <tr>
   *    <td><code><b>x any any</b></code></td>
   *    <td><code>false</code></td>
   *    <td><code>false</code></td>
   *    <td><code>false</code></td>
   *    <td><code>false</code></td>
   *    <td><code>false</code></td>
   *    <td><code>false</code></td>
   *    <td><code>false</code></td>
   *    <td><code>false</code></td>
   *   </tr>
   *   <tr>
   *    <td><code><b>x != a</b></code></td>
   *    <td><code>if a = b</code></td>
   *    <td><code>false</code></td>
   *    <td><code>false</code></td>
   *    <td><code>false</code></td>
   *    <td><code>false</code></td>
   *    <td><code>false</code></td>
   *    <td><code>false</code></td>
   *    <td><code>false</code></td>
   *   </tr>
   *  </table>
   * <li>For example:
   * <pre>
   * x = 10 conflicts with x = 20
   * x = 10 conflicts with x < 10
   * x = 10 conflicts with x > 10
   * x = ten conflicts with x *= ing but not with x *= n
   * x = ten conflicts with x =* anti but not with x =* t
   * x = ten conflicts with x ** rst but not with x ** e
   * x = 10 conflicts with x != 10
   *
   * x < 10 conflicts with x = 10
   * x < 10 conflicts with x > 10
   * 
   * x > 10 conflicts with x = 10
   * x > 10 conflicts with x < 10
   *
   * x *= ing conflicts with x *= abc, but not with x *= g
   * x =* anti conflicts with x =* pre, but not with x =* ant
   * </pre>
   * </ul>
   */
  bool conflicting( const constraint& c1, const constraint& c2 );

  /** @brief Procedure to determine if constraint <code>c1</code>
   * covers constraint <code>c2</code>.
   *
   * <code>c1</code> covers <code>c2</code> if <code>c2</code> <em>implies</em>
   * <code>c1</code>.
   *
   * Constraint <code>c1</code> <b>always</b> covers <code>c2</code>
   * if its operator is siena::any_id (regardless of type) or if
   * it is exactly the same as <code>c2</code>.
   *
   * <p>
   *
   * Otherwise, constraints <b>never</b> cover each other if:
   * <ul>
   * <li> They are they same object (addresses are equal).
   * <li> Their names are not equal.
   * <li> Their types are not equal.
   * </ul>
   * <p>
   * Otherwise, <code>c1</code>(rows) covers <code>c2</code>(columns) according to this table:
   *  <table>
   *   <tr>
   *    <td>&nbsp;</td>
   *    <td><code><b>x = b</b></code></td>
   *    <td><code><b>x < b</b></code></td>
   *    <td><code><b>x > b</b></code></td>
   *    <td><code><b>x *= b</b></code></td>
   *    <td><code><b>x =* b</b></code></td>
   *    <td><code><b>x ** b</b></code></td>
   *    <td><code><b>x any any</b></code></td>
   *    <td><code><b>x != b</b></code></td>
   *   </tr>
   *   <tr>
   *    <td><code><b>x = a</b></code></td>
   *    <td><code>false</code></td>
   *    <td><code>false</code></td>
   *    <td><code>false</code></td>
   *    <td><code>false</code></td>
   *    <td><code>false</code></td>
   *    <td><code>false</code></td>
   *    <td><code>false</code></td>
   *    <td><code>false</code></td>
   *   </tr>
   *   <tr>
   *    <td><code><b>x < a</b></code></td>
   *    <td><code>if b < a</code></td>
   *    <td><code>if b < a</code></td>
   *    <td><code>false</code></td>
   *    <td><code>false</code></td>
   *    <td><code>false</code></td>
   *    <td><code>false</code></td>
   *    <td><code>false</code></td>
   *    <td><code>false</code></td>
   *   </tr>
   *   <tr>
   *    <td><code><b>x > a</b></code></td>
   *    <td><code>if b > a</code></td>
   *    <td><code>false</code></td>
   *    <td><code>if b > a</Code></td>
   *    <td><code>false</code></td>
   *    <td><code>false</code></td>
   *    <td><code>false</code></td>
   *    <td><code>false</code></td>
   *    <td><code>false</code></td>
   *   </tr>
   *   <tr>
   *    <td><code><b>x *= a</b></code></td>
   *    <td><code>if a *= b</code></td>
   *    <td><code>false</code></td>
   *    <td><code>false</code></td>
   *    <td><code>false</code></td>
   *    <td><code>false</code></td>
   *    <td><code>false</code></td>
   *    <td><code>false</code></td>
   *    <td><code>false</code></td>
   *   </tr>
   *   <tr>
   *    <td><code><b>x =* a</b></code></td>
   *    <td><code>if a =* b</code></td>
   *    <td><code>false</code></td>
   *    <td><code>false</code></td>
   *    <td><code>false</code></td>
   *    <td><code>false</code></td>
   *    <td><code>false</code></td>
   *    <td><code>false</code></td>
   *    <td><code>false</code></td>
   *   </tr>
   *   <tr>
   *    <td><code><b>x ** a</b></code></td>
   *    <td><code>if a ** b</code></td>
   *    <td><code>false</code></td>
   *    <td><code>false</code></td>
   *    <td><code>false</code></td>
   *    <td><code>false</code></td>
   *    <td><code>false</code></td>
   *    <td><code>false</code></td>
   *    <td><code>false</code></td>
   *   </tr>
   *   <tr>
   *    <td><code><b>x any any</b></code></td>
   *    <td><code>true</code></td>
   *    <td><code>true</code></td>
   *    <td><code>true</code></td>
   *    <td><code>true</code></td>
   *    <td><code>true</code></td>
   *    <td><code>true</code></td>
   *    <td><code>true</code></td>
   *    <td><code>true</code></td>
   *   </tr>
   *   <tr>
   *    <td><code><b>x != a</b></code></td>
   *    <td><code>false</code></td>
   *    <td><code>false</code></td>
   *    <td><code>false</code></td>
   *    <td><code>false</code></td>
   *    <td><code>false</code></td>
   *    <td><code>false</code></td>
   *    <td><code>false</code></td>
   *    <td><code>false</code></td>
   *   </tr>
   *  </table>
   * <p>For example:
   * <pre>
   * x < 14 covers x < 11
   * x > 5 covers x > 12
   * x < 10 covers x = 8
   * x > 10 covers x = 15
   * x > 10 does not cover x = 9
   * x < 1 does not cover x = 2
   * x *= ing covers x = jumping
   * x =* anti covers x = antidote
   * x ** ym covers x = oxymoron
   * </pre>
   */
  bool covers( const constraint& c1, const constraint& c2 );

  /** @brief Procedure to determine if a filter covers another one.
   *
   * For <code>f1</code> to cover <code>f2</code>, every constraint
   * in <code>f1</code> must cover a constraint in <code>f2</code>.
   * <p>
   * For example:
   * <pre>
   * x < 14 && y > 5 covers x < 11 && y > 12
   * x < 10 && y > 10 covers x = 8 && y = 15
   * x < 10 && y > 10 && z = 4 does not cover x = 8 && y = 15
   * </pre>
   */
  bool covers( const filter& f1, const filter& f2 );

  /** @brief Procedure to determine if a predicate covers another one.
   *
   * For <code>p1</code> to cover <code>p2</code>, every filter in
   * <code>p2</code> must be covered by some filter in
   * <code>p1</code>.
   */
  bool covers( const predicate& p1, const predicate& p2 );
};

#endif
